package com.sun.beans.util;

import java.lang.ref.WeakReference;
import java.lang.ref.SoftReference;
import java.lang.ref.Reference;
import java.util.Objects;
import java.lang.ref.ReferenceQueue;

public abstract class Cache<K, V>
{
    private static final int MAXIMUM_CAPACITY = 1073741824;
    private final boolean identity;
    private final Cache.Kind keyKind;
    private final Cache.Kind valueKind;
    private final ReferenceQueue<Object> queue;
    private volatile Cache.CacheEntry<K, V>[] table;
    private int threshold;
    private int size;
    
    public abstract V create(K p0);
    
    public Cache(Cache.Kind kind, Cache.Kind kind2) {
        this(kind, kind2, false);
    }
    
    public Cache(Cache.Kind kind, Cache.Kind kind2, boolean identity) {
        this.queue = new ReferenceQueue();
        this.table = this.newTable(8);
        this.threshold = 6;
        Objects.requireNonNull(kind, "keyKind");
        Objects.requireNonNull(kind2, "valueKind");
        this.keyKind = kind;
        this.valueKind = kind2;
        this.identity = identity;
    }
    
    public final V get(K obj) {
        Objects.requireNonNull(obj, "key");
        this.removeStaleEntries();
        int hash = this.hash(obj);
        Cache.CacheEntry[] table = this.table;
        Object entryValue = this.getEntryValue(obj, hash, table[index(hash, table)]);
        if (entryValue != null) {
            return entryValue;
        }
        synchronized (this.queue) {
            Object entryValue2 = this.getEntryValue(obj, hash, this.table[index(hash, this.table)]);
            if (entryValue2 != null) {
                return entryValue2;
            }
            Object create = this.create(obj);
            Objects.requireNonNull(create, "value");
            int index = index(hash, this.table);
            this.table[index] = new Cache.CacheEntry(hash, obj, create, this.table[index]);
            if (++this.size >= this.threshold) {
                if (this.table.length == 1073741824) {
                    this.threshold = Integer.MAX_VALUE;
                }
                else {
                    this.removeStaleEntries();
                    Cache.CacheEntry[] table2 = this.newTable(this.table.length << 1);
                    this.transfer(this.table, table2);
                    if (this.size >= this.threshold / 2) {
                        this.table = table2;
                        this.threshold <<= 1;
                    }
                    else {
                        this.transfer(table2, this.table);
                    }
                    this.removeStaleEntries();
                }
            }
            return create;
        }
    }
    
    public final void remove(K k) {
        if (k != null) {
            synchronized (this.queue) {
                this.removeStaleEntries();
                int hash = this.hash(k);
                int index = index(hash, this.table);
                Cache.CacheEntry<Object, Object> cacheEntry = this.table[index];
                while (cacheEntry != null) {
                    Cache.CacheEntry<Object, Object> access$100 = cacheEntry.next;
                    if (Cache.CacheEntry.access$200(cacheEntry, hash, k)) {
                        if (cacheEntry == cacheEntry) {
                            this.table[index] = access$100;
                        }
                        else {
                            cacheEntry.next = access$100;
                        }
                        Cache.CacheEntry.access$300(cacheEntry);
                        break;
                    }
                    else {
                        cacheEntry = access$100;
                    }
                }
            }
        }
    }
    
    public final void clear() {
        synchronized (this.queue) {
            for (int length = this.table.length; 0 < length--; this.table[length] = null) {
                Cache.CacheEntry<Object, Object> access$100;
                for (Cache.CacheEntry cacheEntry = this.table[length]; cacheEntry != null; cacheEntry = access$100) {
                    access$100 = cacheEntry.next;
                    Cache.CacheEntry.access$300(cacheEntry);
                }
            }
            while (null != this.queue.poll()) {}
        }
    }
    
    private int hash(Object o) {
        if (this.identity) {
            int identityHashCode = System.identityHashCode(o);
            return (identityHashCode << 1) - (identityHashCode << 8);
        }
        int hashCode = o.hashCode();
        int n = hashCode ^ (hashCode >>> 20 ^ hashCode >>> 12);
        return n ^ n >>> 7 ^ n >>> 4;
    }
    
    private static int index(int n, Object[] array) {
        return n & array.length - 1;
    }
    
    private Cache.CacheEntry<K, V>[] newTable(int n) {
        return (Cache.CacheEntry[])new Cache.CacheEntry[n];
    }
    
    private V getEntryValue(K k, int n, Cache.CacheEntry<K, V> access$100) {
        while (access$100 != null) {
            if (Cache.CacheEntry.access$200(access$100, n, k)) {
                return access$100.value.getReferent();
            }
            access$100 = access$100.next;
        }
        return null;
    }
    
    private void removeStaleEntries() {
        Reference<Object> reference = this.queue.poll();
        if (reference != null) {
            while (true) {
                try {
                    if ((reference instanceof Cache.Ref)) {
                        Cache.CacheEntry<Object, Object> cacheEntry = (Cache.CacheEntry)((Cache.Ref)reference).getOwner();
                        if (cacheEntry != null) {
                            int index = index(cacheEntry.hash, this.table);
                            Cache.CacheEntry<Object, Object> cacheEntry2 = this.table[index];
                            while (cacheEntry2 != null) {
                                Cache.CacheEntry<Object, Object> access$100 = cacheEntry2.next;
                                if (cacheEntry2 == cacheEntry) {
                                    if (cacheEntry2 == cacheEntry2) {
                                        this.table[index] = access$100;
                                    }
                                    else {
                                        cacheEntry2.next = access$100;
                                    }
                                    Cache.CacheEntry.access$300(cacheEntry2);
                                    break;
                                }
                                else {
                                    cacheEntry2 = access$100;
                                }
                            }
                        }
                    }
                    reference = this.queue.poll();
                    continue;
                    break;
                }
                finally {
                }
            }
        }
    }
    
    private void transfer(Cache.CacheEntry<K, V>[] array, Cache.CacheEntry<K, V>[] array2) {
        int length = array.length;
        while (0 < length--) {
            Cache.CacheEntry<K, V> cacheEntry = array[length];
            array[length] = null;
            while (cacheEntry != null) {
                Cache.CacheEntry<Object, Object> access$100 = cacheEntry.next;
                if (cacheEntry.key.isStale() || cacheEntry.value.isStale()) {
                    Cache.CacheEntry.access$300(cacheEntry);
                }
                else {
                    int index = index(cacheEntry.hash, array2);
                    cacheEntry.next = array2[index];
                    array2[index] = cacheEntry;
                }
                cacheEntry = access$100;
            }
        }
    }
    
    private final class CacheEntry<K, V>
    {
        private final int hash;
        private final Cache.Ref<K> key;
        private final Cache.Ref<V> value;
        private volatile Cache.CacheEntry<K, V> next;
        
        private CacheEntry(int hash, K k, V v, Cache.CacheEntry<K, V> next) {
            this.hash = hash;
            this.key = Cache.this.keyKind.create(this, k, Cache.this.queue);
            this.value = Cache.this.valueKind.create(this, v, Cache.this.queue);
            this.next = next;
        }
        
        private boolean matches(int n, Object obj) {
            if (this.hash != n) {
                return false;
            }
            Object referent = this.key.getReferent();
            return referent == obj || (!Cache.this.identity && referent != null && referent.equals(obj));
        }
        
        private void unlink() {
            this.next = null;
            this.key.removeOwner();
            this.value.removeOwner();
            Cache.this.size--;
        }
        
        static boolean access$200(Cache.CacheEntry cacheEntry, int n, Object o) {
            return cacheEntry.matches(n, o);
        }
        
        static void access$300(Cache.CacheEntry cacheEntry) {
            cacheEntry.unlink();
        }
    }
    
    public enum Kind
    {
        STRONG {
            @Override
             <T> Cache.Ref<T> create(Object o, T t, ReferenceQueue<? super T> referenceQueue) {
                return new Cache.Kind.Strong(o, t);
            }
        }, 
        SOFT {
            @Override
             <T> Cache.Ref<T> create(Object o, T t, ReferenceQueue<? super T> referenceQueue) {
                return (t == null) ? new Cache.Kind.Strong(o, t) : new Cache.Kind.Soft(o, t, referenceQueue);
            }
        }, 
        WEAK {
            @Override
             <T> Cache.Ref<T> create(Object o, T t, ReferenceQueue<? super T> referenceQueue) {
                return (t == null) ? new Cache.Kind.Strong(o, t) : new Cache.Kind.Weak(o, t, referenceQueue);
            }
        };
        
        abstract <T> Cache.Ref<T> create(Object p0, T p1, ReferenceQueue<? super T> p2);
        
        private static final class Strong<T> implements Cache.Ref<T>
        {
            private Object owner;
            private final T referent;
            
            private Strong(Object owner, T referent) {
                this.owner = owner;
                this.referent = referent;
            }
            
            @Override
            public Object getOwner() {
                return this.owner;
            }
            
            @Override
            public T getReferent() {
                return this.referent;
            }
            
            @Override
            public boolean isStale() {
                return false;
            }
            
            @Override
            public void removeOwner() {
                this.owner = null;
            }
        }
        
        private static final class Soft<T> extends SoftReference<T> implements Cache.Ref<T>
        {
            private Object owner;
            
            private Soft(Object owner, T referent, ReferenceQueue<? super T> q) {
                super(referent, q);
                this.owner = owner;
            }
            
            @Override
            public Object getOwner() {
                return this.owner;
            }
            
            @Override
            public T getReferent() {
                return this.get();
            }
            
            @Override
            public boolean isStale() {
                return null == this.get();
            }
            
            @Override
            public void removeOwner() {
                this.owner = null;
            }
        }
        
        private static final class Weak<T> extends WeakReference<T> implements Cache.Ref<T>
        {
            private Object owner;
            
            private Weak(Object owner, T referent, ReferenceQueue<? super T> q) {
                super(referent, q);
                this.owner = owner;
            }
            
            @Override
            public Object getOwner() {
                return this.owner;
            }
            
            @Override
            public T getReferent() {
                return this.get();
            }
            
            @Override
            public boolean isStale() {
                return null == this.get();
            }
            
            @Override
            public void removeOwner() {
                this.owner = null;
            }
        }
    }
    
    private interface Ref<T>
    {
        public abstract Object getOwner();
        
        public abstract T getReferent();
        
        public abstract boolean isStale();
        
        public abstract void removeOwner();
    }
}
